%% Po.sl -*- mode: slang; mode: fold -*-
%%
%% po.sl -- an Emacs-like Jed editing mode for portable object files (*.po)
%% 
%% Copyright    : http://www.fsf.org/copyleft/gpl.html
%% Author       : Morten Bo Johansen <mojo@mbjnet dk>
%%
%% Requirements: bufutils.sl (http://jedmodes.sourceforge.net/)
%%
%% Installation: Copy this file and po.hlp to a directory in your
%% jed library path.
%%
%% In your .jedrc insert the following lines:
%%
%% autoload ("po_mode", "po");
%% add_mode_for_extension ("po", "po");
%% add_mode_for_extension ("po", "pot");
%%
%% For the rest, please refer to the po.hlp file for the ins and outs
%% of this mode. If you copied it to the same place as this file, it
%% should be available in po_mode by typing '?'.
%%

% _debug_info=1;_traceback=1;_slangtrace=1;

%{{{ variables

%% Name and email address of you, the translator
custom_variable ("Translator", sprintf ("%s <%s>", get_realname (), get_emailaddress ()));
%% Name and email address of your language team
custom_variable ("Language_Team", "Danish <dansk@klid.dk>"); % language and address of team
% Character encoding of po-file
custom_variable ("Charset", "ISO-8859-1");
%% If po-file has plural forms: how many plural forms does your language
%% have? See /usr/share/doc/gettext-doc/gettext_10.html#SEC1.
custom_variable ("Nplurals", "2");
%% See /usr/share/doc/gettext-doc/gettext_10.html#SEC1
custom_variable ("Plural", "n != 1");
% Set the revision date when saving: 0 = never, 1 = ask, 2 = always
custom_variable ("Revise_Date", 2);
%% How much of the original text of the entry being translated should be 
%% displayed in the top window?
%% 0 = only the msgid itself in narrowed fold
%% 1 = msgid with all comments in narrowed fold
%% 2 = window split in two equal halves with msgid highlighted
custom_variable ("Msgid_Context", 0);
%% Name and address of the Free Translation Project's mail robot
custom_variable ("Po_Robot", "The Free Translation Project <translation@iro.umontreal.ca>");
%% Program to use for spell checking; use ispell or aspell
custom_variable ("Spell_Prg", "ispell");
%% Language dictionary for ispell/aspell, e.g. "deutsch" for German
custom_variable ("Spell_Dict", "dansk");

static variable
  mode = "PO",
  bg = "",
  edit_buf = "",
  plural_str_number = "",
  msgid_search_expr = "^[#~ ]*msgid[_ ]*",
  msgstr_search_expr = "^[#~ ]*msgstr[[ ]*",
  Use_Newline_Literals = NULL,
  Translation_Status = 0;

%}}}

%{{{ delimitation

%% Some handling of obsolete entries is accomodated
static define is_po_string_delimiter ()
{
   bol_skip_white ();
   orelse
     { looking_at ("#\n") }
     { looking_at ("#,") }
     { looking_at ("#.") }
     { looking_at ("#:") }
     { looking_at ("# ") }
     { looking_at ("msgid ") }
     { looking_at ("#~ #") }
     { looking_at ("#~ msgid ") }
     { eolp () };
}

%% Mark msgstr and msgid, excluding comments. No matter where you are
%% on the entry and whether entries are separated by comments or blank
%% space or not, the strings should be correctly marked.
define mark_strings ()
{
   while (not (is_po_string_delimiter ())) () = up (1);
   !if (re_fsearch (msgid_search_expr)) () = re_bsearch (msgid_search_expr);
   push_mark (); () = down (1);
   while (not (is_po_string_delimiter ())) () = down (1);
   call ("narrow_to_region");
   bob ();
}

%% Like mark_strings, but including comments
static define po_mark_entry ()
{
   mark_strings ();
   call ("widen_region");
   () = up (1);
   while (is_po_string_delimiter () and not bobp ()) () = up (1);
   while (not is_po_string_delimiter ()) () = down (1);
   bol_skip_white ();
   if (eolp ()) () = down (1);
   push_mark ();
   while (is_po_string_delimiter ()) () = down (1);
   while (not (is_po_string_delimiter ())) () = down (1);
   call ("narrow_to_region");
   bob ();
}

%% The character prefixes for translator comments
static define is_translator_comment ()
{
   bol ();
   re_looking_at ("#[ \t\n]+");
}

%% The character prefixes for source comments, auto comments and flags
static define is_other_comment ()
{
   bol ();
   re_looking_at ("#[:.,]+");
}

%% Determine whether an entry is untranslated
static define is_untranslated ()
{
   () = ffind ("\"");
   !if (looking_at ("\"\"")) return 0;
   push_spot ();
   skip_chars ("\"\t\n ");
   bol ();
   orelse
     { is_po_string_delimiter () }
     { looking_at ("msgstr") }; % plural msgstr
   pop_spot ();
}

static define is_translated ()
{
   not is_untranslated ();
}

static define is_msgid_plural ()
{
   push_spot (); mark_strings ();
   bol_fsearch ("msgid_plural");
   call ("widen_region"); pop_spot ();
}

static define mark_translator_comment ()
{
   po_mark_entry (); push_mark ();
   while (is_translator_comment ()) () = down (1);
   call ("widen_region");
}

define mark_msgid ()
{
   if (Msgid_Context == 1) % include comments
     {
        po_mark_entry (); push_mark ();
        () = re_fsearch (msgstr_search_expr);
        call ("widen_region");
        return;
     }
   mark_strings ();
   if (plural_str_number == "0") % variable is set in po_edit ()
     () = fsearch ("\"");
   else
     {
        () = fsearch ("msgid_plural");
        () = ffind ("\"");
     }
   !if (looking_at ("\"\""))
     {
        if (Msgid_Context == 2)
          {
             push_visible_mark ();
             eol ();
          }
        else  push_mark_eol ();
     }
   else
     {
        () = down (1);
        if (Msgid_Context == 2)
          push_visible_mark ();
        else
          push_mark ();
        () = re_fsearch (msgstr_search_expr);
        () = bsearch ("\"");
        () = right (1);
     }
   call ("widen_region");
}

static define mark_msgstr ()
{
   if (is_msgid_plural ())
     {
        if (plural_str_number == "0")
          {
             () = bol_fsearch ("msgstr"); 
             () = ffind ("\"");
             push_mark ();
             () = bol_fsearch ("msgstr"); 
             bskip_chars ("\n");
             call ("widen_region");
             return;
          }
        else
          {
             () = re_fsearch (msgstr_search_expr + string (plural_str_number));
             () = ffind ("\"");
             push_mark ();
             while (down (1))
               if (looking_at ("#") or looking_at ("msgstr")) break;
             bskip_chars ("\n");
          }
     }
   else
     {
        mark_strings ();
        () = re_fsearch (msgstr_search_expr);
        () = fsearch ("\"");
        push_mark_eob ();
     }
   bskip_chars ("\n");
   call ("widen_region");
}

%}}}

%{{{ po-header

%% The default header from emacs' po-mode
static define insert_po_header ()
{
   set_readonly (0);
   bob ();
   while (is_translator_comment ()) () = down (1);
   insert
     ("# SOME DESCRIPTIVE TITLE.\n" +
      "# Copyright (C) YEAR Free Software Foundation, Inc.\n" +
      "# FIRST AUTHOR <EMAIL@ADDRESS>, YEAR.\n" +
      "#\n" +
      "#, fuzzy\n" +
      "msgid \"\"\n" +
      "msgstr \"\"\n" +
      "\"Project-Id-Version: PACKAGE VERSION\\n\"\n" +
      "\"PO-Revision-Date: YEAR-MO-DA HO:MI +ZONE\\n\"\n" +
      "\"Last-Translator: FULL NAME <EMAIL@ADDRESS>\\n\"\n" +
      "\"Language-Team: LANGUAGE <LL@li.org>\\n\"\n" +
      "\"MIME-Version: 1.0\\n\"\n" +
      "\"Content-Type: text/plain; charset=CHARSET\\n\"\n" +
      "\"Content-Transfer-Encoding: 8bit\\n\"\n\n");

   bob ();
   set_readonly (1);
}

%% Check for default headers
static define old_headers_found ()
{
   variable cs = CASE_SEARCH;

   EXIT_BLOCK
     {
        CASE_SEARCH = cs;
        call ("widen_region");
        pop_spot ();
     }

   CASE_SEARCH = 1;

   push_spot_bob ();
   mark_strings ();
   orelse
     { fsearch (": PACKAGE") }
     { fsearch (" VERSION") }
     { fsearch (": LANGUAGE") }
     { fsearch ("LL@li.org") };
}

%% Adjust the po-header to user's settings.
define replace_headers ()
{
   set_readonly (0);
   push_spot_bob ();
   mark_strings ();
   replace ("PACKAGE VERSION", whatbuf ());
   replace ("FULL NAME <EMAIL@ADDRESS>", Translator);
   replace ("LANGUAGE <LL@li.org>", Language_Team);
   replace ("charset=CHARSET", strcat ("charset=", Charset));   
   replace ("INTEGER", Nplurals);
   replace ("EXPRESSION", Plural);
   pop_spot ();
   set_readonly (1);
   call ("widen_region");
}

% Set the revision date
define set_po_revision_date ()
{
   variable gm, gmhour, lm, lmhour, offset;
   variable tm, day, month, year;
   tm = localtime (_time ());
   gm = (gmtime (_time ()));
   gmhour = gm.tm_hour;
   lm = (localtime (_time ()));
   lmhour = lm.tm_hour;
   offset = (lmhour-gmhour)*100;
   variable t = sprintf ("%d-%02d-%02d %02d:%02d", 1900 + tm.tm_year, 1+tm.tm_mon, tm.tm_mday, tm.tm_hour, tm.tm_min);
   bob ();
   mark_strings ();
   !if (bol_fsearch ("\"PO-Revision-Date"))
     {
        call ("widen_region");
        error ("No date header found");
     }
   () = ffind (" ");
   set_readonly (0);
   del_eol ();
   whitespace (1);
   vinsert ("%s+%04d%s", t, offset, "\\n\"");
   call ("widen_region");
   set_readonly (1);
}

%% Set the revision date when saving
static define set_date_on_save ()
{
   variable mode = get_mode_name ();
   !if ("PO" == mode ()) return;

   push_spot ();

   if (Revise_Date == 1)
     {
        switch (get_mini_response ("May I set the PO Revison date?: [y/n]"))
          { case 'y': set_po_revision_date (); }
          { case 'n': pop_spot (); return; }
     }
   else if (Revise_Date == 2)
     {
        set_po_revision_date ();
     }
   pop_spot ();
}

append_to_hook ("_jed_save_buffer_before_hooks", &set_date_on_save);

%}}}

%{{{ some functions relating to statistics

%% Make all count variables local to the buffer
static define count (i)
{
   switch (i)
     { case "t": get_blocal_var ("translated"); }
     { case "u": get_blocal_var ("untranslated"); }
     { case "f": get_blocal_var ("fuzzy"); }
     { case "o": get_blocal_var ("obsolete"); }
     { case "t+": set_blocal_var (get_blocal_var ("translated") + 1, "translated"); }
     { case "t-": set_blocal_var (get_blocal_var ("translated") - 1, "translated"); }
     { case "u+": set_blocal_var (get_blocal_var ("untranslated") + 1, "untranslated"); }
     { case "u-": set_blocal_var (get_blocal_var ("untranslated") - 1, "untranslated"); }
     { case "f+": set_blocal_var (get_blocal_var ("fuzzy") + 1, "fuzzy"); }
     { case "f-": set_blocal_var (get_blocal_var ("fuzzy") - 1, "fuzzy"); }
     { case "o+": set_blocal_var (get_blocal_var ("obsolete") + 1, "obsolete"); }
     { case "o-": set_blocal_var (get_blocal_var ("obsolete") - 1, "obsolete"); }
}

%% The status line that shows the counts for translated, untranslated, fuzzy
%% and obsolete entries
static define set_po_status_line ()
{
   variable t, u, f, o, s;
   
   t = count ("t"); u = count ("u"); f = count ("f"); o = count ("o");
   s = Sprintf ("%dt/%du/%df/%do", t, u, f, o, 4);
   set_status_line (" %b  " + s + "  (%m%a%n%o)  %p   %t", 0);
}

%% The actual counting of the various entries. Untranslated entries that
%% incorrectly are marked fuzzy are counted as untranslated and 
%% not fuzzy entries as they should
static define po_statistics ()
{
   variable t, u, f, o;
   
   create_blocal_var ("translated"); 
   set_blocal_var (0, "translated");
   create_blocal_var ("untranslated"); 
   set_blocal_var (0, "untranslated");
   create_blocal_var ("fuzzy"); 
   set_blocal_var (0, "fuzzy");
   create_blocal_var ("obsolete"); 
   set_blocal_var (0, "obsolete");
   
   t = count ("t"); u = count ("u"); f = count ("f"); o = count ("o");

   ERROR_BLOCK
     {
        pop_spot ();
        _clear_error ();
        error ("Maybe po-file has syntax errors? Try to validate ..");
     }

   if (eobp () and bobp ())
     {
        insert_po_header ();
        return;
     }

   flush ("Counting entries ...");
   push_spot_bob ();
   () = bol_fsearch ("msgstr"); % skip past a fuzzy flag in header
   while (down (1))             % do all the counts
     {
        if (looking_at ("#, fuzzy"))
          {
             push_spot ();
             () = down (1);
             if (looking_at ("#~")) o++;
             else if (andelse { bol_fsearch ("msgstr") } { is_untranslated }) ;
             else f++;
             pop_spot ();
          }
        else if (looking_at ("#~ msgid")) o++;
        if (looking_at ("msgstr"))
          {
             if (is_untranslated ()) u++;
             else t++;
          }
     }
   
   t = t - f; % Translated count doesn't exlude fuzzies so deduct them here
   
   set_blocal_var (t, "translated");
   set_blocal_var (u, "untranslated");
   set_blocal_var (f, "fuzzy");
   set_blocal_var (o, "obsolete");
   
   pop_spot (); 
   set_po_status_line ();
}

%% Show the counts in the message area including total counts and current
%% position
define show_po_statistics ()
{
   variable t, u, f, o, n = 0, cn = 0;
   mark_strings ();
   () = re_fsearch (msgid_search_expr);
   call ("widen_region");
   variable pos = create_user_mark ();
   bob (); 
   () = bol_fsearch ("msgstr");
   while (re_fsearch (msgid_search_expr))
     { 
        n++;
        if (pos == create_user_mark) cn = n; % get the number of current entry
        eol ();
     }
   goto_user_mark (pos);
   po_statistics ();
   t = count ("t"); u = count ("u"); f = count ("f"); o = count ("o");
   vmessage ("%s %d/%d; %d translated, %d untranslated, %d fuzzy, %d obsolete", "Position:", cn, n, t, u, f, o);
}

%}}}

%{{{ navigation

define po_recenter ()
{
   recenter (0);
}

define po_next_entry ()
{
   mark_strings (); eob ();
   call ("widen_region");
   !if (re_fsearch (msgid_search_expr))
     flush ("You are on the last entry.");
}

define po_end_of_entry ()
{
   mark_strings (); eob ();
   call ("widen_region");
}

define po_previous_entry ()
{
   mark_strings ();
   call ("widen_region");
   !if (re_bsearch (msgid_search_expr))
     flush ("You are on the first entry.");
}

%% Find the next untranslated entry and wrap search around buffer
define find_untranslated ()
{
   variable s = create_user_mark ();
   
   if (count ("u") == 0)
     { 
        flush ("No untranslated strings");
        return;
     }
   
   po_end_of_entry ();
   
   while (bol_fsearch ("msgstr"))
     {
        if (is_untranslated)
          { 
             () = bol_bsearch ("msgid");
             return;
          }
     }
   flush ("Wrapping search around buffer");
   bob ();
   while (bol_fsearch ("msgstr"))
     {
        if (is_untranslated)
          { 
             () = bol_bsearch ("msgid");
             return;
          }
     }
   flush ("No untranslated strings");
   goto_user_mark (s);
}

%% Find the previous untranslated msgstr.
define bfind_untranslated ()
{
   if (count ("u") == 0)
     { 
        flush ("No untranslated strings");
        return;
     }
   variable s = create_user_mark ();
   forever
     {
        if (andelse { bol_bsearch ("msgstr") } { is_untranslated } )
          {
             () = bol_bsearch ("msgid");
             return;
          }
        else if (andelse { bol_bsearch ("msgstr") } { is_translated } )
          {
             () = bol_bsearch ("msgid");
             continue;
          }
        else break;
     }
   flush ("No untranslated entries above this point");
   goto_user_mark (s);
}

%Find the next translated entry and wrap search around buffer
define find_translated ()
{
   variable s = create_user_mark ();
   
   if (count ("t") == 0)
     { 
        flush ("No more translated strings");
        return;
     }
   po_end_of_entry ();
   while (down (1))
     {
        if (andelse { looking_at ("msgstr") } { is_translated () } )
          { 
             () = bol_bsearch ("msgid");
             return;
          }
        if (ffind (", fuzzy") or looking_at ("#~")) () = fsearch ("msgstr");
     }
   flush ("Wrapping search around buffer");
   bob (); 
   po_end_of_entry ();
   while (down (1))
     {
        if (andelse { looking_at ("msgstr") } { is_translated () } )
          { 
             () = bol_bsearch ("msgid");
             return;
          }
        if (ffind (", fuzzy") or looking_at ("#~")) () = fsearch ("msgstr");
     }
   flush ("No more translated strings");
   goto_user_mark (s);
}

%% Find the previous translated msgstr.
define bfind_translated ()
{
   if (count ("t") == 0)
     { 
        flush ("No translated strings");
        return;
     }
   while (bol_bsearch ("msgstr"))
     {
        if (is_translated ())
          {
             () = bol_bsearch ("msgid");
             return;
          }
        else () = bol_bsearch ("msgid");
     }
}

%% Find the next entry with fuzzy flag and wrap search around buffer.
define find_fuzzy ()
{
   if (count ("f") == 0)
     { 
        flush ("No fuzzy entries");
        return;
     }
   if (bol_fsearch ("#, fuzzy"))
     {
        () = down (1);
        if (looking_at ("#~")) continue;
        else return;
     }
   flush ("Wrapping search around buffer");
   bob ();
   if (bol_fsearch ("#, fuzzy"))
     {
        () = down (1);
        if (looking_at ("#~")) continue;
        else return;
     }
}

%% Find the previous entry with fuzzy flag.
define bfind_fuzzy ()
{
   if (count ("f") == 0)
     { 
        flush ("No fuzzy entries");
        return;
     }
   if (bol_bsearch ("#, fuzzy")) return;
   flush ("No fuzzy entries above this point");
}

%% Find the next obsolete entry and wrap search around buffer
define find_obsolete ()
{
   if (count ("o") == 0)
     { 
        flush ("No obsolete entries");
        return;
     }
   if (fsearch ("#~ msgid"))
     {
        () = down (1);
        return;
     }
   flush ("Wrapping search around buffer");
   bob ();
   if (fsearch ("#~ msgid"))
     {
        () = down (1);
        return;
     }
}

%% Find the previous obsolete entry.
define bfind_obsolete ()
{
   if (count ("o") == 0)
     { 
        flush ("No obsolete entries");
        return;
     }
   if (bol_bsearch ("#~ msgid")) return;
   flush ("No obsolete entries above this point");
}

%% Find the next entry with a translator comment
define find_translator_comment ()
{
   variable s = create_user_mark ();
   if (re_fsearch ("^#[ \t\n].*"))
     {
        () = re_fsearch (msgid_search_expr);
        return;
     }
   bob ();
   po_next_entry ();
   flush ("Wrapped search around buffer..");
   if (re_fsearch ("^#[ \t\n].*"))
     {
        () = re_fsearch (msgid_search_expr);
        return;
     }
   flush ("No translator comments");
   goto_user_mark (s);
}

%}}}

%{{{ modification of entries outside of po_edit buffer

static define del_msgstr ()
{
   mark_msgstr ();
   del_region ();
}

%% Mark entry obsolete.
static define obsolete_entry ()
{
   call ("widen_region");
   !if (get_y_or_n ("Make entry obsolete")) return;
   if (eobp ()) return;
   po_mark_entry ();
   
   do
     {
        if (eolp () and bolp ()) break;
        insert ("#~ ");
     }
   while (down (1));
   
   if (bsearch ("#, fuzzy") and count ("f") > 0) count ("f-");
   count ("o+");
}

%% Flag translation as "fuzzy", or obsolete entry if already fuzzy or
%% untranslated
define fuzzyize ()
{
   ERROR_BLOCK { set_readonly (1); }
   
   EXIT_BLOCK
     {
        call ("widen_region"); 
        pop_spot ();
        set_po_status_line ();
        set_readonly (1);
     }
   
   bol ();
   push_spot (); po_mark_entry ();
   if (looking_at ("#~")) return;
   set_readonly (0);
   if (bol_fsearch ("#, fuzzy")) obsolete_entry ();
   else if (bol_fsearch ("msgstr") and is_untranslated ()) obsolete_entry ();
   else if (re_bsearch ("^#,.*format"))
     {
        skip_chars ("#, ");
        insert ("fuzzy, ");
        if (count ("t") > 0) count ("t-");
        count ("f+");
     }
   else
     {
        () = bol_bsearch ("msgid");
        insert ("#, fuzzy\n");
        if (count ("t") > 0) count ("t-");
        count ("f+");
     }
}

define remove_fuzzy_flag ()
{
   push_spot ();
   EXIT_BLOCK 
     { 
        call ("widen_region"); 
        pop_spot ();
        set_po_status_line ();
        set_readonly (1);
     }
   
   po_mark_entry ();
   !if (bol_fsearch ("#, fuzzy")) return;
   set_readonly (0);
   bob ();
   if (re_fsearch ("^#, fuzzy[ \t]*$"))
     { 
        delete_line ();
        if (count ("f") > 0) count ("f-");
        count ("t+");
     }
   else if (re_fsearch ("^#, fuzzy,.*format"))
     {
        skip_chars ("#, "); push_mark (); () = ffind (" "); () = right (1);
        del_region ();
        if (count ("f") > 0) count ("f-");
        count ("t+");
     }
}

define del_translator_comment ()
{
   set_readonly (0);
   mark_translator_comment ();
   del_region ();
   set_readonly (1);
}

%% Delete an obsolete entry
define del_obsolete ()
{
   po_mark_entry ();
   !if (fsearch ("#~"))
     {
        call ("widen_region");
        error ("Not an obsolete entry");
     }
   else
     {
        mark_buffer ();
        set_readonly (0);
        del_region ();
        call ("widen_region");
        bol_skip_white ();
        if (eolp () and not eobp ()) () = del ("\n");
        set_readonly (1);
        if (count ("o") > 0) count ("o-");
        set_po_status_line ();
     }
}

define unobsolete_entry ()
{
   push_spot ();
   po_mark_entry ();
   set_readonly (0);
   while (looking_at ("#~ "))
     {
        do
          {
             push_mark ();
             () = ffind (" "); 
             skip_chars (" \t");
             del_region ();
          }
        while (down (1));
     }
   if (count ("o") > 0) count ("o-");
   count ("f+");
   set_po_status_line ();
   call ("widen_region");
   pop_spot ();
   set_readonly (1);
}

%% Remove contents of msgstr (leaving it unstranslated)
define cut_msgstr ()
{
   mark_strings ();
   call ("widen_region");
   () = re_fsearch (msgstr_search_expr);
   if (is_untranslated ()) return;
   set_readonly (0);
   po_mark_entry ();
   if (re_fsearch ("^#, fuzzy[ \t]*$"))
     { 
        delete_line ();
        if (count ("f") > 0) count ("f-");
     }
   else if (re_fsearch ("^#, fuzzy,.*format"))
     {
        skip_chars ("#, "); push_mark (); () = ffind (" "); () = right (1);
        del_region ();
        if (count ("f") > 0) count ("f-");
     }
   call ("widen_region");
   del_msgstr ();
   insert ("\"\"");
   set_readonly (1);
   if (count ("t") > 0) count ("t-");
   count ("u+");
   set_po_status_line ();
}

%% Copy contents of msgid to msgstr
define copy_msgid_to_msgstr ()
{
   mark_strings ();
   () = re_fsearch (msgstr_search_expr);
   call ("widen_region");
   if (is_untranslated ())
     {
        if (count ("u") > 0) count ("u-");
        count ("t+");
        set_po_status_line ();
     }
   else
     {
        !if (get_y_or_n ("Overwrite previous translation")) return;
     }
   mark_msgid ();
   call ("narrow_to_region");
   mark_buffer ();
   variable n = what_line ();
   variable msgid = bufsubstr ();
   call ("widen_region");
   set_readonly (0);
   del_msgstr ();
   push_spot ();
   if (n > 1)
     {
        insert ("\"\"");
        () = down (1);
        insert (msgid);
        insert_char ('\n');
     }
   else
     insert (msgid);
   pop_spot ();
   set_readonly (1);
}

define copy_msgstr ()
{
   create_blocal_var ("msgstr_copy");
   push_spot ();
   mark_msgstr ();
   set_blocal_var (bufsubstr (), "msgstr_copy");
   pop_spot ();
   flush ("msgstr copied");
}

define insert_msgstr ()
{
   !if (blocal_var_exists ("msgstr_copy"))
     error ("Nothing to insert");

   variable str = get_blocal_var ("msgstr_copy");
   mark_strings ();
   () = bol_fsearch ("msgstr");
   call ("widen_region");
   if (is_untranslated ())
     {
        if (count ("u") > 0) count ("u-");
        count ("t+");
        set_po_status_line ();
     }
   else
     {
        !if (get_y_or_n ("Overwrite previous translation")) return;
     }
   call ("widen_region");
   set_readonly (0);
   del_msgstr ();
   insert (str);
   set_readonly (1);
}

define po_undo ()
{
   set_readonly (0);
   call ("undo");
   po_statistics ();
   set_readonly (1);
}

%}}}

%{{{ po_edit buffer

%% Mode for the editing buffer.
static define po_edit_mode ()
{
   set_mode ("po_edit", 0);
   use_keymap ("po_edit");
   set_abbrev_mode (1);
   if (abbrev_table_p ("po_edit")) use_abbrev_table ("po_edit");
   run_mode_hooks ("po_edit_mode_hook");
   % set_buffer_hook ("newline_indent_hook", "insert_lbreak");
   set_buffer_modified_flag (0);
   set_buffer_undo (1);
}

define decreasewin ()
{
   call ("other_window");
   enlargewin ();
   call ("other_window");
}

%% Toggle termination of lines in msgstr with "\n" on and off
define toggle_newline_literals ()
{
   variable s = Null_String;
   if (blocal_var_exists (s))
     Use_Newline_Literals = get_blocal_var (s);
   else
     create_blocal_var (s);
   if (Use_Newline_Literals == NULL)
     {
        Use_Newline_Literals = 1;
        flush ("Newline characters will be represented by \"\\n\"");
        set_blocal_var (Use_Newline_Literals, s);
     }
   else
     {
        Use_Newline_Literals = 0;
        flush ("No newline literals");
        set_blocal_var (NULL, s);
     }
}

%% Strip double quotes on line ends from msgstr before editing
static define strip_quotes ()
{
   bob ();
   if (looking_at ("\"\"")) delete_line ();
   if (bobp () and eobp ()) return;
   do
     {
        bol (); del ();
        eol (); () = bfind ("\""); del ();
     }
   while (down (1) and not eobp ());
}

%% De-escape double quotes before editing
static define de_esc_quotes ()
{
   if (eobp () and bobp ()) return;
   bob ();
   replace ("\\\"", "\"");
}

static define strip_newline_literals ()
{
   if (fsearch ("\\n"))
     replace ("\\n", Null_String);
}

%% No blank lines in po_edit buffer after msgstr
static define trim_blank_lines ()
{
   eob ();
   bskip_chars ("\n");
   if (eobp ()) return;
   do
     {
        del ();
     }
   while (not eobp ());
}

%% Escape double quotes before merging edit buffer with po-file
static define esc_quotes ()
{
   bob ();
   replace ("\"", "\\\"");
}

%% Surround all msgstr lines in double quotes after finishing editing
static define add_quotes ()
{
   bob ();
   do
     {
        bol (); insert ("\"");
        eol (); insert ("\"");
     }
   while (down (1));
}

static define insert_newline_literals ()
{
   bob ();
   replace ("\n", "\\n\n");
}

define insert_lbreak ()
{
   insert ("\\n\n");
}

define restore_po_buffer_window ()
{
   call ("other_window");
   call ("one_window");
   call ("widen_region");
}

%% A split window is created with the msgid to be translated in the top window
%% and the translation to be filled in is in the bottom window.
define po_edit ()
{
   variable msgstr = "", n;
   edit_buf = "*" + whatbuf () + "*";
   if (is_msgid_plural ())
     {
        bol ();
        if (looking_at ("msgstr[0]")) 
          plural_str_number = "0";
        else if (looking_at ("msgstr[1]"))
          plural_str_number = "1";
        else
          plural_str_number = read_mini ("Enter number of msgstr to edit:", "", "");
     }
   mark_msgstr ();
   msgstr = bufsubstr ();
   mark_msgid ();
   if (Msgid_Context != 2)
     {
        call ("narrow_to_region");
        n = what_line ();
     }
   recenter (1); % make sure msgid window is the top window
   pop2buf (edit_buf);
   call ("other_window");
   if (Msgid_Context == 2)
     {
        call ("exchange");
        recenter (5); % approx. center msgid in top window
     }
   else
     {
        if (n <= 13)
          {
             autoload ("window_set_rows", "bufutils");
             window_set_rows (n + 1);
          }
        bob ();
     }
   call ("other_window");
   bob ();
   insert (msgstr);
   strip_quotes ();
   de_esc_quotes ();
   if (Use_Newline_Literals == 1)
     strip_newline_literals ();

   if (eobp () and bobp ())
     Translation_Status = 0;
   else
     Translation_Status = 1;

   po_edit_mode ();
}

define del_editbuf ()
{
   set_buffer_modified_flag (0);
   delbuf (edit_buf);
   restore_po_buffer_window ();
   if (Msgid_Context == 2) pop_mark (0);
}

%% The finished translation is read back into the po-file
define po_end_edit ()
{
   !if (buffer_modified)
     {
        del_editbuf ();
        po_next_entry ();
        return;
     }

   trim_blank_lines ();

   if (Use_Newline_Literals == 1)
     insert_newline_literals ();

   esc_quotes ();
   add_quotes ();
   variable buf = "", n, len;
   mark_buffer ();
   n = what_line ();
   buf = bufsubstr ();
   len = strlen (buf);
   set_buffer_modified_flag (0);
   restore_po_buffer_window ();
   if (Msgid_Context == 2) pop_mark (0);
   set_readonly (0);
   del_msgstr ();
   if (len >= 72 or n > 1)
     {
        insert ("\"\"\n");
        insbuf (edit_buf);
     }
   else insbuf (edit_buf);
   delbuf (edit_buf);
   () = re_bsearch (msgstr_search_expr);
   if (Translation_Status == 0 and not (is_untranslated ()))
     {
        count ("t+");
        if (count ("u") > 0) count ("u-");
     }
   else if (Translation_Status == 1 and (is_untranslated ()))
     {
        remove_fuzzy_flag ();
        if (count ("t") > 0) count ("t-");
        count ("u+");
     }
   set_po_status_line ();
   set_readonly (1);
   po_next_entry ();
}

%}}}

%{{{ translator comments

static define po_comment_mode ()
{
   set_mode ("po_comment", 0);
   use_keymap ("po_comment");
   set_buffer_undo (1);
}

static define strip_comments ()
{
   bob ();
   while (is_translator_comment)
     {
        do
          {
             bol (); push_mark ();
             skip_chars ("# \t");
             del_region ();
          }
        while (down (1));
     }
}

static define add_comments ()
{
   if (eobp () and bobp ()) return;
   bob ();
   do
     {
        bol (); insert ("# ");
     }
   while (down (1));
   eob ();
   insert_char ('\n');
}

%% Insert a translator comment or edit an existing one
define po_edit_comment ()
{
   edit_buf = "*" + whatbuf () + "*";
   variable comment = "";
   push_spot ();
   mark_translator_comment ();
   comment = bufsubstr ();
   po_mark_entry (); eob ();
   variable n = what_line ();
   recenter (1);
   pop2buf (edit_buf);
   insert (comment);
   strip_comments ();
   call ("other_window");
   if (n <= 13)
     window_set_rows (n);
   bob ();
   call ("other_window");
   bob ();
   po_comment_mode ();
   set_buffer_modified_flag (0);
}

%% This happens after we leave the po_comment buffer
define po_end_edit_comment ()
{
   !if (buffer_modified)
     {
        delbuf (edit_buf);
        restore_po_buffer_window ();
        pop_spot ();
        return;
     }
   trim_blank_lines ();
   variable buf = "";
   mark_buffer ();
   buf = bufsubstr ();
   add_comments ();
   set_buffer_modified_flag (0);
   restore_po_buffer_window ();
   del_translator_comment ();
   set_readonly (0);
   insbuf (edit_buf);
   delbuf (edit_buf);
   pop_spot ();
   set_readonly (1);
}

%}}}

%{{{ miscellaneous functions

%% Exit po_mode
define edit_whole_buffer ()
{
   no_mode ();
   use_keymap ("global");
   toggle_readonly ();
   call ("redraw");
   set_status_line ("", 0);
   flush ("Use \"M-x po_mode\" to return");
}

%% Search for a file in $PATH
static define prg_found_in_path (file)
{
   variable path = getenv ("PATH");
   path = (path, ) = strreplace (path, ":", ",", strlen (path));
   return search_path_for_file (path, file);
}

%% Validate po-file with the gettext msgfmt program
define po_validate_command ()
{
   if (NULL == prg_found_in_path ("msgfmt"))
     error ("msgfmt program not found in $PATH");
   compile ("msgfmt --statistics -c -v -o /dev/null " + whatbuf());
}

%% Compile the po-file so it can be used by the mother program
define po_compile ()
{
   if (NULL == prg_found_in_path ("msgfmt"))
     error ("msgfmt program not found in $PATH");
   variable dir = dircat (getenv ("HOME"), Null_String);
   variable str = read_mini ("Compile command: ", "", "msgfmt" +
                             " -o " + dir + path_sans_extname (whatbuf()) +
                             ".mo " + whatbuf());
   if (strlen (str)) compile (str);
}

%% Use the gettext program 'msgunfmt' to decompile a *.mo file and read it
%% into the editor
define po_decompile ()
{
   if (NULL == prg_found_in_path ("msgunfmt"))
     error ("msgunfmt program not found in $PATH");
   
   variable mo_file = read_with_completion ("*.mo-file:", "", "", 'f');
   variable po_file = dircat (getenv ("HOME"), strcat (path_sans_extname (path_basename (mo_file)) + ".po"));
   
   if (file_status (po_file) and (not get_y_or_n ("File exists, overwrite")))
     return;
   else
     {
        system (strcat ("msgunfmt ", mo_file, " > ", po_file));
        find_file (po_file);
     }
}

%% Set the path to source files
define po_set_srcpath ()
{
   create_blocal_var ("Source_Dir");
   variable s = read_with_completion ("Source dir:", "", "", 'f');
   set_blocal_var (s, "Source_Dir");
}

%% Get source references and return them as a formatted string
static define po_mark_src_str ()
{
   variable str = "";
   push_spot ();
   po_mark_entry ();
   !if (bol_fsearch ("#:"))
     {
        call ("widen_region");
        pop_spot ();
        error ("No source reference");
     }
   push_mark ();
   while (looking_at ("#:")) () = down (1);
   str = bufsubstr ();
   str = (str, ) = strreplace(str, "\n", " ", strlen (str));
   str = strtrim ((str, ) = strreplace (str, "#: ", Null_String, strlen (str)));
   call ("widen_region");
   pop_spot ();
   return (str);
}

%% Pop up windows containing files from source references for viewing
define view_source ()
{
   variable str, buf, ln, line, srcfile;
   variable cnt = 1;

   str = po_mark_src_str ();
   ln = length (strchop (str, ' ', 0));

   foreach (strchop (str, ' ', 0))
     {
        $0 = ();
        $1 = strchop ($0, ':', 0);
        buf = $1[0];
        line = $1[1];
        !if (blocal_var_exists ("Source_Dir"))
          {
             flush ("Path to source files not set, use 'S' to set it");
             return;
          }
        srcfile = path_concat (get_blocal_var ("Source_Dir"), buf);
        !if (1 == file_status (srcfile))
          {
             vmessage ("Source file %s not found (use 'S' to set the path)",
                       srcfile);
             return;
          }
        variable oldbuf = pop2buf_whatbuf (buf);
        () = insert_file (srcfile);
        set_buffer_modified_flag (0);
        set_readonly (1);
        goto_line (integer (line));
        local_setkey ("close_file", "q");
        c_mode ();
        call ("one_window");

        if (ln > 1)
          {
             vmessage ("%s %d %s %d %s", "Source reference", cnt, "of", ln,
                       "(space cycles, 'q' closes window, other key to scroll)");
             while (cnt == ln)
               {
                  vmessage ("%s %d %s %d %s", "Source reference", cnt, "of",
                            ln, "(space or 'q' closes window, other key to scroll)");
                  break;
               }
          }
        else vmessage ("%s %d %s %d %s", "Source reference", cnt, "of", ln,
                       "(space or 'q' closes window, other key to scroll)");

        update (1);

        variable ch = getkey ();

        switch (ch)
          { ch == ' ': delbuf (buf); sw2buf (oldbuf); cnt++; }
          { ch == 'q': delbuf (buf); return; }
          { ch != ' ' or 'q': flush ("'q' closes window"); return; }
     }
}

%% Use msgcat to wrap/unwrap entries
define wrap_entries (unwrap)
{
   if (NULL == prg_found_in_path ("msgcat"))
     error ("msgcat program not found in $PATH");

   variable tmpfile = make_tmp_file ("/tmp/po_wrap");
   mark_buffer (); 
   () = dupmark ();
   if (unwrap)
     {
        flush ("Unwrapping entries ...");
        pipe_region (strcat ("msgcat --no-wrap -", " > ", tmpfile));
     }
   else
     {
        flush ("Wrapping entries ...");
        pipe_region (strcat ("msgcat -", " > ", tmpfile));
     }
   set_readonly (0);
   del_region ();
   () = insert_file (tmpfile);
   () = delete_file (tmpfile);
   set_readonly (1);
}

%% Send po-file to the translation robot or language team
define mail_po_file ()
{
   variable buf = whatbuf ();
   variable cmd;
   variable file = strcat (buf + ".gz");
   variable tmpfile = make_tmp_file ("/tmp/pomail");
   variable u, f, o;

   u = count ("u"); f = count ("f"); o = count ("o");

   if (old_headers_found ())
     error ("You need to edit the 'Language_Team' and/or 'Project-Id-Version' fields");

   !if (NULL == prg_found_in_path ("msgcat"))
     {
        flush ("Validating file ...");
        push_spot ();
        mark_buffer ();
        !if (0 == pipe_region (sprintf ("msgcat - >/dev/null 2>&1")))
          {
             pop_spot ();
             error ("File has parse errors");
          }
        pop_spot ();
     }
   else
     {
        flush ("Gettext utilities seemingly not installed, could not validate");
        update (1); sleep (2);
     }
   
   if (orelse { count ("u") != 0 } { count ("f") != 0 } { count ("o") != 0 })
     !if (get_y_or_n ("Unprocessed entries remain, continue")) return;

   if (get_y_or_n ("Include po-file in message"))
     {
        variable incl_pofile = "yes";
        variable str = "";
        push_spot ();
        mark_buffer ();
        str = bufsubstr ();
        pop_spot ();
     }

   switch (get_mini_response ("Mail your team [t] or translation robot [r] ?"))
     { case 't':
        mail ();
        insert (Language_Team);
        () = bol_fsearch ("Subject:"); eol ();
        insert (read_mini ("Subject:", path_sans_extname (file), ""));
        eob (); push_spot ();
        insert_char ('\n');
        if (__is_initialized (&incl_pofile))
          insert (str);
     }
     { case 'r':
        mail ();
        insert (Po_Robot);
        () = bol_fsearch ("Subject:");  eol ();
        % mandatory subject for robot
        vinsert ("%s %s", "TP-Robot", path_sans_extname (file));
        eob (); push_spot ();
        insert_char ('\n');
        % gzip and uuencode if uuencode program exists, else incl. as
        % raw text
        if (NULL == prg_found_in_path ("uuencode") and __is_initialized (&incl_pofile))
          {
             flush ("uuencode program not found ...including as raw text.");
             insert (str);
          }
        else
          {
             if (__is_initialized (&incl_pofile))
               {
                  if (NULL == prg_found_in_path ("msgcat"))
                    {
                       flush ("msgcat not found; wrapped entries will not be unwrapped");
                       cmd = "gzip -9 | uuencode -m";
                    }
                  else
                    cmd = "msgcat - --no-wrap | gzip -9 | uuencode -m";
                  setbuf (buf);
                  push_spot ();
                  mark_buffer ();
                  () = pipe_region (sprintf ("%s > %s %s", cmd, tmpfile, file));
                  pop_spot ();
                  setbuf ("*mail*");
                  () = insert_file (tmpfile);
               }
          }
     }
   pop_spot ();
   () = delete_file (tmpfile);
   __uninitialize (&incl_pofile);
}

%% Display help window
define show_help ()
{
   variable file = expand_jedlib_file ("po.hlp");
   !if (1 == file_status (file))
     {
        flush ("The file 'po.hlp' was not found in Jed's library path");
        return;
     }
   () = find_file (file);
   most_mode ();
   flush ("Scroll down for more. 'q' closes help window");
}

%% Send a bug report
define reportbug ()
{
   mail ();
   insert ("Morten Bo Johansen <mojo@mbjnet.dk>");
   () = bol_fsearch ("Subject:"); eol ();
}

%% Interactively spell check msgstrs with ispell or aspell. If Jed is run 
%% under X, the interactive spelling process will be opened in a separate
%% terminal window
define po_spellcheck ()
{
   if (count ("t") == 0)
     {
        flush ("Nothing to spell check");
        return;
     }

   variable TermPrg;
   
   !if (NULL == prg_found_in_path ("rxvt"))
     TermPrg = "rxvt -e ";
   else
     TermPrg = "xterm -e ";
   
   if (NULL == prg_found_in_path ("pospell"))
     {
        flush ("spellutils not installed or $PATH is incomplete");
        return;
     }

   if (NULL == prg_found_in_path (Spell_Prg))
     {
        vmessage ("%s %s", Spell_Prg, "not found in $PATH");
        return;
     }

   variable tmpfile, aspell_cmd, ispell_cmd, line;
   tmpfile = make_tmp_file ("/tmp/spellfile");
   
   aspell_cmd = strcat ("pospell -n ", tmpfile, " -p ", Spell_Prg, " -- -c -d ", Spell_Dict, " %f");
   ispell_cmd = strcat ("pospell -n ", tmpfile, " -p ", Spell_Prg, " -- -x -C -d ", Spell_Dict, " %f");
   
   line = what_line ();

   ERROR_BLOCK
     {
        goto_line (line);
        recenter (0);
        set_readonly (1);
        () = delete_file (tmpfile);
        () = run_shell_cmd (sprintf ("%s %s", "rm >/dev/null 2>&1", "/tmp/newsbody*"));
     }
   
   mark_buffer ();
   () = write_region_to_file (sprintf ("%s", tmpfile));
   
   if (Spell_Prg == "ispell")
     {
        if (is_defined ("x_server_vendor"))
          system (strcat (TermPrg, ispell_cmd));
        else 
          vrun_program ("%s", ispell_cmd);
     }
   if (Spell_Prg == "aspell")
     {
        if (is_defined ("x_server_vendor"))
          system (strcat (TermPrg, aspell_cmd));
        else 
          vrun_program ("%s", aspell_cmd);
     }
   set_readonly (0);
   mark_buffer ();
   del_region ();
   () = insert_file (tmpfile);

   EXECUTE_ERROR_BLOCK;
}

%}}}

%{{{ dfa syntax and colors

#ifdef HAS_DFA_SYNTAX
%%% DFA_CACHE_BEGIN %%%
create_syntax_table (mode);
static define setup_dfa_callback (mode)
{
   (, bg) = get_color ("normal"); % that dfa backgr. fits with user's color

   add_color_object ("strings");
   add_color_object ("srccomment");
   add_color_object ("flagcomment");
   add_color_object ("autocomment");
   add_color_object ("usercomment");
   add_color_object ("lbreak");
   add_color_object ("begblank");
   add_color_object ("endblank");

   if (is_defined ("x_server_vendor"))
     {
        set_color ("strings", "magenta", bg);
        set_color ("srccomment", "gray", bg);
        set_color ("flagcomment", "red", bg);
        set_color ("autocomment", "gray", bg);
        set_color ("usercomment", "black", "green");
        set_color ("lbreak","blue", bg);
        set_color ("begblank","black", "cyan");
        set_color ("endblank","black", "cyan");
     }
   else
     {
        set_color ("strings", "magenta", bg);
        set_color ("srccomment", "gray", bg);
        set_color ("flagcomment", "red", bg);
        set_color ("autocomment", "gray", bg);
        set_color ("usercomment", "black", "green");
        set_color ("lbreak","blue", bg);
        set_color ("begblank","black", "cyan");
        set_color ("endblank","black", "cyan");
     }

   dfa_enable_highlight_cache ("po.dfa", mode);
   dfa_define_highlight_rule ("^msg(id|str)", "strings", mode);
   dfa_define_highlight_rule ("^#[,]+.*", "flagcomment", mode);
   dfa_define_highlight_rule ("^#[:]+.*", "srccomment", mode);
   dfa_define_highlight_rule ("^#[\\.]+.*", "autocomment", mode);
   dfa_define_highlight_rule ("^#[ \t]+.*", "usercomment", mode);
   dfa_define_highlight_rule ("[ \t]*\"[ \t]+", "begblank", mode);
   dfa_define_highlight_rule ("[ \t]+\\\\?n?\"$", "endblank", mode);
   dfa_define_highlight_rule ("\\\\n", "lbreak", mode);
   dfa_build_highlight_table (mode);
}
dfa_set_init_callback (&setup_dfa_callback, mode);
%%% DFA_CACHE_END %%%
#endif

%}}}

%{{{ Menu

static define po_init_menu (menu)
{
   menu_append_popup (menu, "&Navigate");
   $0 = menu + ".&Navigate";
     {
        menu_append_item ($0, "Next Entry", "po_next_entry");
        menu_append_item ($0, "Previous Entry", "po_previous_entry");
        menu_append_item ($0, "Next Untranslated", "find_untranslated");
        menu_append_item ($0, "Previous Untranslated", "bfind_untranslated");
        menu_append_item ($0, "Next Translated", "find_translated");
        menu_append_item ($0, "Previous Translated", "bfind_translated");
        menu_append_item ($0, "Next Fuzzy", "find_fuzzy");
        menu_append_item ($0, "Previous Fuzzy", "bfind_fuzzy");
        menu_append_item ($0, "Next Obsolete", "find_obsolete");
        menu_append_item ($0, "Previous Obsolete", "bfind_obsolete");
        menu_append_item ($0, "Next Translator Comment", "find_translator_comment");
     }
   menu_append_popup (menu, "&Modify");
   $1 = menu + ".&Modify";
     {
        menu_append_item ($1, "Undo",  "po_undo");
        menu_append_item ($1, "Edit Msgstr", "po_edit");
        menu_append_item ($1, "Replace Headers", "replace_headers");
        menu_append_item ($1, "Copy Msgstr", "copy_msgstr");
        menu_append_item ($1, "Insert Msgstr", "insert_msgstr");
        menu_append_item ($1, "Copy Msgid To Msgstr", "copy_msgid_to_msgstr");
        menu_append_item ($1, "Cut Msgstr", "cut_msgstr");
        menu_append_item ($1, "Flag Entry As Fuzzy", "fuzzyize");
        menu_append_item ($1, "Remove Fuzzy Flag", "remove_fuzzy_flag");
        menu_append_item ($1, "Delete Obsolete Entry", "del_obsolete");
        menu_append_item ($1, "Remove Obsolete Flag", "unobsolete_entry");
        menu_append_item ($1, "Edit Comment", "edit_comment");
        menu_append_item ($1, "Delete Comment", "del_translator_comment");
        menu_append_item ($1, "Edit Entire File", "edit_whole_buffer");
        menu_append_item ($1, "Spell Check", "po_spellcheck");
        menu_append_item ($1, "Wrap Entries", "wrap_entries (0)");
        menu_append_item ($1, "Unwrap Entries", "wrap_entries (1)");
     }
   menu_append_popup (menu, "&Po-Edit Buffer Functions");
   $2 = menu + ".&Po-Edit Buffer Functions";
     {
        menu_append_item ($2, "Finish Editing           TAB", "po_end_edit");
        menu_append_item ($2, "Enlarge Msgstr Window    ESC +", "enlargewin");
        menu_append_item ($2, "Decrease Msgstr Window   ESC -", "decreasewin");
        menu_append_item ($2, "Insert Newline Literal   ESC-ENTER", "insert_lbreak");
        menu_append_item ($2, "Reject Changes           Ctrl-C", "del_editbuf");
     }
   menu_append_item (menu, "Show Help",  "show_help");
   menu_append_item (menu, "Compile *.po -> *.mo", "po_compile");
   menu_append_item (menu, "Decompile *.mo -> *.po", "po_decompile");
   menu_append_item (menu, "Validate", "po_validate_command");
   menu_append_item (menu, "Statistics", "show_po_statistics");
   menu_append_item (menu, "Toggle Newline Literals", "toggle_newline_literals");
   menu_append_item (menu, "Set Source Path", "po_set_srcpath");
   menu_append_item (menu, "View Source Reference", "view_source");
   menu_append_item (menu, "Send Po-file to Robot/Team", "mail_po_file");
   menu_append_item (menu, "Send a Bug Report", "reportbug");
}

%}}}

%{{{ Keymaps
   
require ("keydefs");

!if (keymap_p ("po_comment"))
{
   make_keymap ("po_comment");
   definekey ("enlargewin", "^[-", "po_comment");      %  esc +
   definekey ("decreasewin", "^[/", "po_comment");     %  esc -
   if (NULL != getenv ("DISPLAY"))
     {
        definekey ("enlargewin", "^[+", "po_comment");
        definekey ("decreasewin", "^[-", "po_comment");
     }
   definekey ("po_end_edit_comment", "^I", "po_comment"); % tab
   definekey ("del_editbuf", "^C", "po_comment");
}

!if (keymap_p ("po_edit"))
{
   make_keymap ("po_edit");
   definekey ("insert_lbreak", "^[^M", "po_edit");
   definekey ("po_end_edit", "^I", "po_edit");
   definekey ("enlargewin", "^[-", "po_edit");
   definekey ("decreasewin", "^[/", "po_edit");
   definekey ("del_editbuf", "^C", "po_edit");
   if (NULL != getenv ("DISPLAY"))
     {
        definekey ("enlargewin", "^[+", "po_edit");
        definekey ("decreasewin", "^[-", "po_edit");
     }
}

!if (keymap_p (mode))
{
   make_keymap (mode);
   definekey ("page_up", "b", mode);
   definekey ("po_recenter", ",", mode);
   definekey ("po_undo", "_", mode);
   definekey ("po_edit_comment", "#", mode);
   definekey ("show_po_statistics", "=", mode);
   definekey ("show_help", "?", mode);
   definekey ("find_fuzzy", "f", mode);
   definekey ("bfind_fuzzy", "F", mode);
   definekey ("po_next_entry", "n", mode);
   definekey ("po_next_entry", " ", mode);
   definekey ("po_previous_entry", "p", mode);
   definekey ("po_compile", "C", mode);
   definekey ("po_decompile", "D", mode);
   definekey ("copy_msgstr", "w", mode);
   definekey ("insert_msgstr", "y", mode);
   definekey ("edit_whole_buffer", "E", mode);
   definekey ("mail_po_file", "M", mode);
   definekey ("toggle_newline_literals", "N", mode);
   definekey ("find_obsolete", "o", mode);
   definekey ("bfind_obsolete", "O", mode);
   definekey ("view_source", "s", mode);
   definekey ("po_set_srcpath", "S", mode);
   definekey ("find_translated", "t", mode);
   definekey ("bfind_translated", "T", mode);
   definekey ("find_untranslated", "u", mode);
   definekey ("bfind_untranslated", "U", mode);
   definekey ("po_validate_command", "V", mode);
   definekey ("po_edit", "^M", mode);                         % enter
   definekey ("po_edit", "^I", mode);                         % tab
   definekey ("del_obsolete", Key_Del, mode);              % del
   definekey ("fuzzyize", Key_BS, mode);                     % backspace
   definekey ("remove_fuzzy_flag", Key_Alt_BS, mode);
   definekey ("find_translator_comment", "\e#", mode);
   definekey_reserved ("del_translator_comment", "#", mode);
   definekey_reserved ("copy_msgid_to_msgstr", "c", mode);
   definekey_reserved ("replace_headers", "h", mode);
   definekey_reserved ("cut_msgstr", "k", mode);
   definekey_reserved ("unobsolete_entry", "o", mode);
   definekey_reserved ("po_spellcheck", "s", mode);
   definekey_reserved ("wrap_entries (1)", "w-", mode);
   definekey_reserved ("wrap_entries (0)", "w+", mode);
}

%}}}

%{{{ Mode definition

public define po_mode ()
{
   set_mode (mode, 0);
   set_readonly (1);
   use_keymap (mode);
   use_syntax_table (mode);
   mode_set_mode_info (mode, "init_mode_menu", &po_init_menu);
   run_mode_hooks ("po_mode_hook");
   po_statistics ();
   message ("F10 -> Mode, gives you access to menu functions. Type '?' for help");
}

provide (mode);

%}}}
